/* Connection.pmod
 *
 * Hller reda p en uppkoppling till lyskomservern.
 */

#define WERROR werror

class _module_value
{
  object lock;
  object socket;
  object ctx;
  object lexer;
  object parser;
  string to_write;

  object(Misc.Queue) requests; /* Anrop som inte har kats upp
				* fr behandling n. */
  object next_request;
  int pending_requests; /* Anrop som brjat behandlats, och
			 * inte har besvarats n */
  int is_syncing;       /* Gr inga nya anrop frrn kn r tmd */
  int is_dying;         /* Kasta alla svar p golvet */

  
  int message_written;
  constant message = "LysKOM\n";

  string login_name;
  
  object session;

  void start_dying()
  {
    WERROR("connection->start_dying()\n");
    object key = lock->lock();
    if (session)
      session->delete();
    is_dying = 1;
    socket->close();
    to_write = 0;
  }

  void close_callback()
  {
    start_dying();
  }

#if 0
  void die()
  {
    WERROR("connection->die\n");
    if (session)
      session->close();
    socket->close();
  }
#endif
  
  void write_callback()
  {
    object key = lock->lock();
    if (is_dying)
      {
	WERROR("Connection->write_callback(): Dying...\n");
	return;
      }
    int written = socket->write(to_write);
    if (written <= 0) 
      start_dying();
    WERROR(sprintf("Wrote '%s'.\n", to_write[..written]));
    to_write = to_write[written..];
    if (!strlen(to_write))
      socket->set_write_callback(0);
  }

  /* Anropas med uppkopplingen lst */
  void may_process_requests()
  {
    WERROR(sprintf("Connection->may_process_requests:\n"
		   "  is_syncing = %d, pending_requests = %d\n",
		   is_syncing, pending_requests));
    if (is_syncing && pending_requests)
      return;
    if (next_request)
      {
	WERROR("Connection->may_process_requests: enqueuing next_request\n");
	is_syncing = next_request->need_sync_after;

	pending_requests++;
	ctx->enqueue_request(next_request);
	next_request = 0;
      }
    while (!(is_syncing && pending_requests) && requests->size())
      {
	object req = requests->get();
	// WERROR(sprintf("Connection->may_process_requests: request %O\n",
	// 	       indices(req)));
	if (req->need_sync_before && pending_requests)
	  {
	    next_request = req;
	    is_syncing = 1;
	    return;
	  }
	is_syncing = req->need_sync_after;
	pending_requests++;
	ctx->enqueue_request(req);
      }
  }
  
  /* Ska anropas med uppkopplingen lst. */
  void send(string response)
  {
    if (is_dying)
      {
	WERROR("Connection->send(): Dying...\n");
	return;
      }
    to_write += response;
    socket->set_write_callback(write_callback);
  }

  void answer(string response)
  {
    /* Anropas typiskt p slutet av en requests process() metod */
    object key = lock->lock();
    send(response);
    pending_requests--;
    may_process_requests();
  }
  
  void send_async(string msg)
  {
    object key = lock->lock();
    send(msg);
  }
  
  void handle_raw_request(array(object) tokens)
  {
    werror(sprintf("Got: %O\n",
		   Array.map(tokens, lambda(object t)
			     { return ({ t->token, t->value }); } )));
    array a = Requests.make_request(tokens);
    if (!a)
      send(protocol_A.protocol_error_to_A());
    
    else if (a[0])
      {
	object req = a[1];
	req->set_context(session, this_object());
	object key = lock->lock();
	WERROR("Connection->handle_raw_request: enqueuing request\n");
	requests->put(req);

	may_process_requests();
	WERROR("Connection->handle_raw_request: considered request\n");
      }
    else
      send(protocol_A.error_to_A(@a[1..]));
  }

  void read_callback(mixed id, string s)
  {
    lexer->add_data(s);
    array r;
    while (r=parser->get_request())
      handle_raw_request(r);
  }

  void handle_connect_3(mixed id)
  {
    WERROR("connection->handle_connect_3\n");
    int written = socket->write(message[message_written..]);
    if (written <= 0)
      {
	werror("connection->print_hello: Write failed.\n");
	start_dying();
      }
    message_written += written;
    if (message_written == strlen(message))
      {
	parser = protocol_A.Dumb_Parser(lexer, 1);
	socket->set_write_callback(0);
	socket->set_read_callback(read_callback);
	session = Session(login_name, send_async);
	array r;
	while( (r=parser->get_request()) )
	  handle_raw_request(r);
      }
  }
    
  void handle_connect_2(object t)
  {
    WERROR(sprintf("connection->handle_connect_2: %O\n", t));
    if (t->token != "string")
      start_dying();
    login_name = t->value;
    socket->set_read_callback(0);
    socket->set_write_callback(handle_connect_3);
  }
  
  void handle_connect_1(mixed id, string s)
  {
    WERROR(sprintf("connection->handle_connect_1: '%s'\n", s));
    if (!strlen(s) || (s[0] != 'A'))
      start_dying();
    lexer->add_data(s[1..]);
    object t = lexer->get_token();
    if (t)
      handle_connect_2(t);
    else
      socket->set_read_callback(lambda(mixed id, string s) {
	lexer->add_data(s);
	object t = lexer->get_token();
	if (t)
	  handle_connect_2(t);
      } );
  }
        
  void create(object f, object c)
  {
    lock = Thread.Mutex();
    requests = Misc.Queue();
    socket = f;
    ctx = c;
    lexer = protocol_A.Lexer();
    // parser = protocol_A.Dumb_Parser();
    message_written = 0;
    to_write = "";
    socket->set_nonblocking(handle_connect_1, 0, close_callback);
  }
}

  
