/* Misc.pmod
 *
 * Diverse anvndbara funktioner.
 */

/* Returnera indexet p n i a.
 * Om inte n finns, returnera indexet till nrmast lgre tal.
 */
int binary_search(int n, array a, void|int start)
{
  int lower = start;
  int upper = sizeof(a);
  int middle;

  if (!sizeof(a) || (n < a[start]))
    return -1;
      
  while (upper > lower)
    {
      middle = (upper + lower) / 2;
      if ((a[middle] == n) || (lower == middle))
	return middle;
      else if (n < a[middle])
	upper = middle;
      else
	lower = middle;
    }
  return -1;
}

array(int) sorted_insert(array(int) a, int n)
{
  int i = binary_search(n, a);
  if (i<0)
    a = ({ n }) + a;
  else
    a = a[..i] + ({ n }) + a[i+1..];
  return a;
}      

string kom_canonicalize_name(string name)
{
  name = lower_case(name);
  if (Config.query_var("swascii_matching"))
    name = replace(name, ({ "}", "{", "|", "]", "[", "\\"}),
		   ({ "", "", "", "", "", ""}) );
  return name;
}

object parenthesis_re = Regexp("^(.*)"
			       "\\("
			       "[^()]*"
			       "\\)"
			       "(.*)$");

array(string) kom_simplify_name(string name)
{
  name = kom_canonicalize_name(name);
  array(string) a;

  /* Ta bort (nstlade) parantetiserade uttryck */
  while(a = parenthesis_re->split(name))
    name = a * "";

  /* Erstt alla whitespace med mellanslag */
  name = replace(name, ({ "\t", "\r", "\n" }), ({ " ", " ", " " }) );

  /* Dela upp i ord, och ta bort tomma ord. */
  return (name / " ") - ({ "" });
}

array(int) pack_array(array(int) a)
{
  int size = sizeof(a);
  array res = allocate(size);
  if (size)
    {
      res[0] = a[0];
      for(int i = 1; i<size; i++)
	res[i] = a[i] - a[i-1];
    }
  return res;
}

array(int) unpack_array(array(int) a)
{
  int size = sizeof(a);
  array(int) res = allocate(size);
  if (size)
    {
      res[0] = a[0];
      for(int i = 1; i<size; i++)
	res[i] = res[i-1] + a[i];
    }
  return res;
}

class Stack
{
  array stack;
  int p;

  void create()
  {
    p = 0;
    stack = allocate(17);
  }

  int is_empty()
  {
    return p == 0;
  }

  int query_size()
  {
    return p;
  }
  
  mixed pop()
  {
    if (!p)
      throw( ({ "protocol_A.stack->pop: Stack is empty.\n", backtrace() }) );
    p--;
    mixed res = stack[p];
    stack[p] = 0;
    return res;
  }

  array pop_from(int n)
  {
    if (p < n)
      throw( ({ "protocol_A.stack->pop_from: Bad.\n", backtrace() }) );
    array res = stack[n..p-1];
    while(p > n)
      stack[--p] = 0;
    return res;
  }
  
  void push(mixed o)
  {
    if (sizeof(stack) == p)
      stack += allocate(sizeof(stack));
    stack[p++] = o;
  }
}

class Queue
{
  array q;
  int head;
  int tail;

  void create(int|void size)
  {
    q = allocate(size|17);
    head = tail = 0;
  }

  int size() { return tail - head; }

  mixed get()
  {
    if (tail > head)
      {
	mixed res = q[head];
	q[head] = 0;
	head++;
	return res;
      }
    else
      throw( ({"protocol_A.Queue->get: Queue empty.\n", backtrace() }) );
  }

  void put(mixed o)
  {
    if (tail == sizeof(q))
      {
	q = q[head..] + allocate(17 + size());
	tail -= head;
	head = 0;
      }
    q[tail++] = o;
  }
}
