/* protocol_A.pmod
 *
 */

#if 0
#define WERROR werror
#else
#define WERROR(x)
#endif

class Token
{
  /* token r "int", "string", "{", "}", "\n",
   * "*", "%", "=" eller ":".
   *
   * "int" anvnds bde fr heltar och bitstrngar, eftersom
   * det inte syns nn skillnad. */
  string token;
  mixed value;

  void create(string t, mixed|void v)
  {
    token = t;
    value = v;
  }
}

class Lexer
{
  constant MODE_NEW_OBJECT = 2;
  constant MODE_SCANNING_STRING = 3;
  constant MODE_SCANNING_INT = 4;
  constant MODE_SKIP_TO_NEWLINE = 5;
  
  int state;
  string buffer;
  int token_start; /* Position i buffer */
  int token_end;   /* Position efter nuvarande hollerithstrng */
  int to_scan;

  object(Misc.Queue) tokens;

  void create()
  {
    state = MODE_NEW_OBJECT;
    buffer = "";
    to_scan = 0;
    tokens = Misc.Queue();
  }

  object(Token) get_token()
  {
    return (tokens->size()) ? tokens->get() : 0;
  }

  int add_data(string s)
  {
    WERROR(sprintf("Lexer->add_data('%s')\n", s));
    buffer += s;
    while(to_scan < strlen(buffer))
      {
	int c = buffer[to_scan];
	WERROR(sprintf("Lexer->add_data: state = %d, c = '%c'\n", state, c));
	switch(state)
	  {
	  case MODE_NEW_OBJECT:
	    switch(c)
	      {
	      case ' ':
	      case '\t':
	      case '\r':
		to_scan++;
		break;
	      case '0'..'9':
		token_start = to_scan;
		state = MODE_SCANNING_INT;
		to_scan++;
		break;
	      case '{':
	      case '}':
	      case '*':
	      case '%':
	      case ':':
	      case '=':
	      case '\n':
		tokens->put(Token(buffer[to_scan..to_scan]));
		to_scan++;
		break;
		
	      default:
		tokens->put(Token("error", buffer[to_scan]));
		to_scan++;
		state = MODE_SKIP_TO_NEWLINE;
		break;
		  
	      }
	    break;
	  case MODE_SCANNING_INT:
	    if ((c >= '0') && (c <= '9'))
	      {
		to_scan++;
		break;
	      }
	    {
	      string n = buffer[token_start..to_scan-1];
	      if (c == 'H')
		{
		  to_scan++;
		  token_start = to_scan;
		  token_end = to_scan + (int) n;
		  state = MODE_SCANNING_STRING;
		}
	      else
		{
		  tokens->put(Token("int", n));
		  state = MODE_NEW_OBJECT;
		}
	    }
	    break;
	  case MODE_SCANNING_STRING:
	    if (sizeof(buffer) < token_end)
	      {
		to_scan = sizeof(buffer);
		break;
	      }
	    tokens->put(Token("string",
			      buffer[token_start..token_end-1]) );

	    /* Shrink buffer */
	    buffer = buffer[token_end..];
	    to_scan = 0;
	    state = MODE_NEW_OBJECT;
	    break;
	  case MODE_SKIP_TO_NEWLINE:
	    {
	      int i = search(buffer, "\n", to_scan);
	      if (i < 0)
		{
		  buffer = "";
		  to_scan = 0;
		}
	      else
		{
		  buffer = buffer[i+1..];
		  to_scan = 0;
		  state = MODE_NEW_OBJECT;
		  tokens->put(Token("\n"));
		}
	      break;
	    }
	  default:
	    throw( ({ "protocol_A.Lexer->add_data: Internal error\n",
			backtrace() }) );
	    
	  }
      }
  }
}

class Dumb_Parser
{
  constant MODE_START_REQUEST = 1;
  constant MODE_START_RESPONSE = 2;
  constant MODE_OBJECT = 3;
  constant MODE_SKIP_TO_NEWLINE = 5;
  

  object lexer;
  int start_state;
  int state;

  object(Misc.Stack) token_stack; /* Scannade object */
  object(Misc.Stack) mark_stack;  /* Positioner fr startmsvingar */
  
  object(Misc.Queue) parsed_requests;
  
  void create(object l, int is_server)
  {
    lexer = l;
    start_state = is_server ? MODE_START_REQUEST : MODE_START_RESPONSE;
    state = start_state;
    token_stack = Misc.Stack();
    mark_stack = Misc.Stack();
    parsed_requests = Misc.Queue();
  }
  
  void parse_token(object(Token) t)
  {
    WERROR(sprintf("Dumb_Parser->parse_token( %O )\n", t));
    while(1)
      {
	WERROR(sprintf("Dumb_Parser->parse_token: state = %d.\n", state));

	switch(state)
	  {
	  case MODE_START_REQUEST:
	    switch(t->token)
	      {
	      case "\n":
		/* Ignore */
		return;
	      default:
		state = MODE_OBJECT;
		break;
	      }
	    break;
	  case MODE_START_RESPONSE:
	    throw( ({ "Dumb_Parser->parse_token: "
			"MODE_RESPONSE not implemented.\n", backtrace() }) );
	    
	  case MODE_OBJECT:
	    switch(t->token)
	      {
	      case "int":
	      case "string":
		token_stack->push(t);
		break;
	      case "*":
		token_stack->push(Token("empty"));
		break;
	      case "{":
		/* First token that is part of the array */
		mark_stack->push(token_stack->query_size());
		break;
	      case "}":
		/* Pop mark stack */
		if (mark_stack->is_empty())
		  {
		    werror("Unbalanced '}'. Skipping to end of line.\n");
		    state = MODE_SKIP_TO_NEWLINE;
		    break;
		  }
		token_stack->push(Token("array", token_stack
				  ->pop_from(mark_stack->pop())));
		break;
	      case "\n": /* End of request */
		if (!mark_stack->is_empty())
		  {
		    werror("Spurious newline: Ignored\n");
		    break;
		  }
		parsed_requests->put(token_stack->pop_from(0));
		state = MODE_START_REQUEST;
		break;
	      default:
		werror(sprintf("Unexpected token: %O."
			       " Skipping to newline.\n", t));
		state = MODE_SKIP_TO_NEWLINE;
		break;
		  
	      }
	    return;
	  case MODE_SKIP_TO_NEWLINE:
	    switch(t->token)
	      {
	      case "\n":
		state = start_state;
		break;
	      default:
		/* Ignore */
		break;
	      }
	      return;
	  default:
	    throw( ({ "protocol_A.Dumb_Parser->parse_token: Internal error\n",
			backtrace() }) );
	    
	  }
      }
  }

  array get_request()
  {
    object t;
    while (!parsed_requests->size() && (t = lexer->get_token()))
      parse_token(t);
      
    return (parsed_requests->size()) ? parsed_requests->get() : 0;
  }
}

  
string string_to_A(string s)
{
  return sprintf("%dH%s", strlen(s), s);
}

string int_to_A(int n)
{
  return sprintf("%d", n);
}

/* Skriv den minst signifikanta bitten frst. */
string bitstring_to_A(int v, int len)
{
  array res = allocate(len);
  for(int i = 0; i < len; i++)
    {
      res[i] = (v % 2) ? "1" : "0";
      v /= 2;
    }
  return res * "";
}

string time_to_A(int t)
{
  mapping lt = localtime(t);
  return sprintf("%d %d %d %d %d %d %d %d %d",
		 lt->sec,  lt->min,  lt->hour,
		 lt->mday, lt->mon,  lt->year,
		 lt->wday, lt->yday, lt->isdst);
}

int tz;

array parse_time(array a)
{
  if (sizeof(column(a, "token") - ({"int"}) ))
    return 0;
  a = a[0..5] + a[8];
  mapping m = mkmapping( ({ "sec", "min", "hour", "mday",
			      "mon", "year", "isdst"}),
			 @Array.map(column(a, "value"),
				    lambda(string x){return (int) x;}) );
  m->timezone = tz;
  return ({ mktime(m) });
}

string array_to_A_low(array(string) a)
{
  WERROR("array_to_a_low: " + (sizeof(a)
			       ? sprintf("%s { %s }",
					 int_to_A(sizeof(a)),
					 a * " ")
			       : "0 *") + "\n");
  return sizeof(a)
    ? sprintf("%s { %s }", int_to_A(sizeof(a)), a * " ")
    : "0 *";
}

string any_to_A(mixed o)
{
  if (stringp(o))
    return string_to_A(o);
  if (intp(o))
    return int_to_A(o);
  if (arrayp(o))
    return array_to_A_low(Array.map(o, any_to_A));
  if (objectp(o))
    return o->query_protA();
  throw( ({ "protocol_A.any_to_A: Unknown type.\n", backtrace() }) );
}

string answer_to_A(int ref_no, array(string) response)
{
  return "=" + int_to_A(ref_no) + " " + response * " " + "\n";
}

#if 0
string answer_to_A(int ref_no, array args)
{
  return "=" + int_to_A(ref_no) + " "
         + Array.map(args, any_to_A) * " " + "\n";
}
#endif

string async_to_A(int msg, array args)
{
  return ":" + int_to_A(sizeof(args)) + " " + int_to_A(msg)
     + Array.map(args, any_to_A) * " " + "\n";
}

string error_to_A(int ref_no, int error_no, int error_status)
{
  return "%" + Array.map(({ ref_no, error_no, error_status}),
			 int_to_A) * " " + "\n";
}

string protocol_error_to_A()
{
  return "%% protocol error\n";
}

void create()
{
  mapping m = localtime(time());
  if (zero_type(m->time_zone))
    werror("No time zone from localtime!\n");
  else
    tz = m->timezone;
}

  
